# pylint: skip-file
import pytest
import pyomo.environ as pe
from suspect.pyomo import dag_from_pyomo_model
from suspect.dag.tags import LinearExpressionTag


class TestLinearExpressionTag:
    def test_sum_of_monomials(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(range(5))

        m.c = pe.Constraint(expr=2.0 + sum((i + 2)*m.x[i] for i in range(5)) + m.x[0] + 3.0 * m.x[1] + 3.0 >= 0)
        dag = dag_from_pyomo_model(m)
        body = dag.constraints['c'].body
        tag = LinearExpressionTag.from_expr(dag, body)
        x0 = dag.variables['x[0]']
        x1 = dag.variables['x[1]']
        assert tag is not None
        assert tag.constant == 5.0
        assert tag.coefficient(x0) == 3.0
        assert tag.coefficient(x1) == 6.0

    def test_sum_of_variables(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(range(5))

        m.c = pe.Constraint(expr=sum(m.x[i] for i in range(5)) >= 0)
        dag = dag_from_pyomo_model(m)
        body = dag.constraints['c'].body
        tag = LinearExpressionTag.from_expr(dag, body)
        assert tag is not None
        for var in dag.variables.values():
            assert tag.coefficient(var) == 1.0



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from numbers import Number
import pyomo.environ as aml
from pyomo.core.kernel.component_map import ComponentMap
import pyomo.core.expr.numeric_expr as pex
from pyomo.core.expr.numeric_expr import (
    nonpyomo_leaf_types,
    NumericConstant,
)
from pyomo.core.expr.visitor import ExpressionValueVisitor
from suspect.pyomo.util import (
    model_variables,
    model_objectives,
    model_constraints,
    bounds_and_expr,
)
from suspect.pyomo.expr_dict import ExpressionDict
from suspect.float_hash import BTreeFloatHasher
from suspect.dag.dag import ProblemDag
import suspect.dag.expressions as dex


def dag_from_pyomo_model(model, return_component_map=False):
    """Convert the Pyomo ``model`` to SUSPECT DAG.

    Parameters
    ----------
    model : ConcreteModel
        the Pyomo model.
    return_component_map : bool
        if True, also return the component map mapping model nodes to dag nodes.

    Returns
    -------
    ProblemDag
        SUSPECT problem DAG.
    """
    dag = ProblemDag(name=model.name)
    factory = ComponentFactory(dag)
    for omo_var in model_variables(model):
        new_var = factory.variable(omo_var)
        dag.add_variable(new_var)

    for omo_cons in model_constraints(model):
        new_cons = factory.constraint(omo_cons)
        dag.add_constraint(new_cons)

    for omo_obj in model_objectives(model):
        new_obj = factory.objective(omo_obj)
        dag.add_objective(new_obj)

    if return_component_map:
        return dag, factory.component_map
    return dag


class ComponentFactory(object):
    def __init__(self, dag):
        self.dag = dag
        self._components = ExpressionDict(float_hasher=BTreeFloatHasher())
        self.component_map = ComponentMap()
        self._visitor = _ConvertExpressionVisitor(self._components, self.dag, self.component_map)

    def variable(self, omo_var):
        comp = self._components.get(omo_var)
        if comp is not None:
            return comp
        new_var = deepcopy(omo_var)
        self._components[omo_var] = new_var
        self.component_map[omo_var] = new_var
        return new_var

    def constraint(self, omo_cons):
        bounds, expr = bounds_and_expr(omo_cons.expr)
        new_expr = self.expression(expr)
        constraint = dex.Constraint(
            omo_cons.name,
            bounds.lower_bound,
            bounds.upper_bound,
            [new_expr],
        )
        return constraint

    def objective(self, omo_obj):
        if omo_obj.is_minimizing():
            sense = dex.Sense.MINIMIZE
        else:
            sense = dex.Sense.MAXIMIZE

        new_expr = self.expression(omo_obj.expr)
        obj = dex.Objective(
            omo_obj.name, sense=sense, children=[new_expr]
        )
        return obj

    def expression(self, expr):
        return self._visitor.dfs_postorder_stack(expr)


class _ConvertExpressionVisitor(ExpressionValueVisitor):
    def __init__(self, memo, dag, component_map):
        self.memo = memo
        self.dag = dag
        self.component_map = component_map

    def visiting_potential_leaf(self, node):
        if node.__class__ in nonpyomo_leaf_types:
            expr = NumericConstant(float(node))
            expr = self.set(expr, expr)
            return True, expr

        if node.is_constant():
            expr = self.set(expr, expr)
            return True, expr

        if node.is_variable_type():
            var = self.get(node)
            assert var is not None
            return True, var
        return False, None

    def visit(self, node, values):
        if self.get(node) is not None:
            new_expr = self.get(node)
            self.set(node, new_expr)
            return new_expr

        new_expr = node.create_node_with_local_data(tuple(values))
        self.set(node, new_expr)
        return new_expr

    def get(self, expr):
        if isinstance(expr, Number):
            const = aml.NumericConstant(expr)
            return self.get(const)
        else:
            return self.memo[expr]

    def set(self, expr, new_expr):
        self.component_map[expr] = new_expr
        if self.memo.get(expr) is not None:
            return self.memo[expr]
        self.memo[expr] = new_expr
        self.dag.add_vertex(new_expr)
        return new_expr



# Copyright 2019 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Re-import Pyomo expression types."""
from pyomo.core.expr.numeric_expr import (
    NumericConstant,
    NegationExpression,
    PowExpression,
    ProductExpression,
    MonomialTermExpression,
    ReciprocalExpression,
    SumExpression,
    UnaryFunctionExpression,
    AbsExpression,
    LinearExpression,
    nonpyomo_leaf_types,
)

from pyomo.core.base.var import Var

from suspect.dag.expressions import Constraint, Objective, Sense, Domain



# pylint: skip-file
import pytest
from tests.conftest import PlaceholderExpression
from suspect.dag.dag import VerticesList
from hypothesis import given, assume
import hypothesis.strategies as st


class TestVerticesList(object):
    @pytest.mark.parametrize('reverse', [True, False])
    @given(depths=st.lists(st.integers(min_value=0, max_value=100), min_size=1))
    def test_append(self, depths, reverse):
        vl = VerticesList(reverse=reverse)
        for d in depths:
            p = PlaceholderExpression()
            vl.append(p, d)
        assert vl._vertices_depth == sorted(depths, reverse=reverse)

    @pytest.mark.parametrize('reverse', [True, False])
    @given(depths=st.lists(st.integers(min_value=0, max_value=100), min_size=1))
    def test_pop(self, depths, reverse):
        vertices = [PlaceholderExpression() for _ in depths]
        depths_map = {}
        vl = VerticesList(reverse=reverse)
        for i, v in enumerate(vertices):
            depths_map[id(v)] = depths[i]
            vl.append(v, depths[i])

        popped = vl.pop()
        if reverse:
            assert depths_map[id(popped)] == max(depths)
        else:
            assert depths_map[id(popped)] == min(depths)

    def test_bool(self):
        vl = VerticesList()
        assert not vl
        vl.append(PlaceholderExpression(), 1)
        assert vl
        vl.pop()
        assert not vl



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds tightening rules."""
from suspect.expression import UnaryFunctionType
from suspect.interfaces import Rule, UnaryFunctionRule
from suspect.interval import Interval
from suspect.math import almosteq  # pylint: disable=no-name-in-module
from suspect.pyomo.expressions import nonpyomo_leaf_types

MAX_EXPR_CHILDREN = 1000


class ConstraintRule(Rule):
    """Return new bounds for constraint."""
    def apply(self, expr, _bounds):
        bounds = Interval(expr.lower_bound, expr.upper_bound)
        return [bounds]


class SumRule(Rule):
    """Return new bounds for sum."""
    def __init__(self, max_expr_children=None):
        if max_expr_children is None:
            max_expr_children = MAX_EXPR_CHILDREN
        self.max_expr_children = max_expr_children

    def apply(self, expr, bounds):
        if len(expr.args) > self.max_expr_children: # pragma: no cover
            return None
        expr_bound = bounds[expr]
        children_bounds = [
            self._child_bounds(child_idx, child, expr, expr_bound, bounds)
            for child_idx, child in enumerate(expr.args)
        ]
        return children_bounds

    def _child_bounds(self, child_idx, _child, expr, expr_bound, bounds):
        siblings_bound = sum(
            bounds[c]
            for i, c in enumerate(expr.args)
            if i != child_idx
        )
        return expr_bound - siblings_bound


class LinearRule(Rule):
    """Return new bounds for linear expressions."""
    def __init__(self, max_expr_children=None):
        if max_expr_children is None:
            max_expr_children = MAX_EXPR_CHILDREN
        self.max_expr_children = max_expr_children

    def apply(self, expr, bounds):
        if expr.nargs() > self.max_expr_children:  # pragma: no cover
            return None
        expr_bound = bounds[expr]
        const = expr.constant
        children_bounds = [
            self._child_bounds(child_idx, coef, child, expr, const, expr_bound, bounds)
            for child_idx, (coef, child) in enumerate(zip(expr.linear_coefs, expr.linear_vars))
        ]
        return children_bounds

    def _child_bounds(self, child_idx, coef, _child, expr, const, expr_bound, bounds):
        siblings_bound = sum(
            bounds[s] * c
            for i, (c, s) in enumerate(zip(expr.linear_coefs, expr.linear_vars))
            if i != child_idx
        ) + const
        return (expr_bound - siblings_bound) / coef


class QuadraticRule(Rule):
    """Return new bounds for quadratic expressions."""
    def __init__(self, max_expr_children=None):
        if max_expr_children is None:
            max_expr_children = MAX_EXPR_CHILDREN
        self.max_expr_children = max_expr_children

    def apply(self, expr, bounds):
        expr_bound = bounds[expr]
        child_bounds = {}
        if len(expr.terms) > self.max_expr_children:
            return None

        terms = expr.terms
        for term_idx, term in enumerate(terms):
            var1 = term.var1
            var2 = term.var2
            siblings_bound = sum(
                self._term_bound(t, bounds)
                for i, t in enumerate(terms) if i != term_idx
            )
            term_bound = (expr_bound - siblings_bound) / term.coefficient
            if var1 is var2:
                term_bound = term_bound.intersect(Interval(0, None))
                upper_bound = term_bound.sqrt().upper_bound
                new_bound = Interval(-upper_bound, upper_bound)

                if id(var1) in child_bounds:
                    existing = child_bounds[id(var1)]
                    child_bounds[id(var1)] = existing.intersect(new_bound)
                else:
                    child_bounds[id(var1)] = Interval(new_bound.lower_bound, new_bound.upper_bound)

            else:
                new_bound_var1 = term_bound / bounds[var2]
                new_bound_var2 = term_bound / bounds[var1]

                if id(var1) in child_bounds:
                    existing = child_bounds[id(var1)]
                    child_bounds[id(var1)] = existing.intersect(new_bound_var1)
                else:
                    child_bounds[id(var1)] = new_bound_var1

                if id(var2) in child_bounds:
                    existing = child_bounds[id(var2)]
                    child_bounds[id(var2)] = existing.intersect(new_bound_var2)
                else:
                    child_bounds[id(var2)] = new_bound_var2

        return [child_bounds[id(v)] for v in expr.args]

    def _term_bound(self, term, bounds):
        if term.var1 is term.var2:
            return term.coefficient * (bounds[term.var1] ** 2)
        return term.coefficient * bounds[term.var1] * bounds[term.var2]


class PowerRule(Rule):
    """Return new bounds for power expressions."""
    def apply(self, expr, bounds):
        base, expo = expr.args
        if type(expo) not in nonpyomo_leaf_types:
            if not expo.is_constant():
                return None
            expo = expo.value

        if not almosteq(expo, 2):
            return None

        expr_bound = bounds[expr]
        # the bound of a square number is never negative, but check anyway to
        # avoid unexpected crashes.
        if not expr_bound.is_nonnegative():
            return None

        sqrt_bound = expr_bound.sqrt()
        return [
            Interval(-sqrt_bound.upper_bound, sqrt_bound.upper_bound),
            None,
        ]


class _UnaryFunctionRule(UnaryFunctionRule):
    def apply(self, expr, bounds):
        expr_bounds = bounds[expr]
        return [self._child_bounds(expr_bounds)]

    def _child_bounds(self, bounds):
        pass


class _BoundedFunctionRule(_UnaryFunctionRule):
    func_name = None

    def _child_bounds(self, bounds):
        func = getattr(bounds, self.func_name)
        return func.inverse()


class AbsRule(_BoundedFunctionRule):
    """Return new bounds for abs."""
    func_name = 'abs'


class SqrtRule(_BoundedFunctionRule):
    """Return new bounds for sqrt."""
    func_name = 'sqrt'


class ExpRule(_BoundedFunctionRule):
    """Return new bounds for exp."""
    func_name = 'exp'


class LogRule(_BoundedFunctionRule):
    """Return new bounds for log."""
    func_name = 'log'


class _UnboundedFunctionRule(_UnaryFunctionRule):
    def _child_bounds(self, bounds):
        return Interval(None, None)


class SinRule(_UnboundedFunctionRule):
    """Return new bounds for sin."""
    func_type = UnaryFunctionType.Sin


class CosRule(_UnboundedFunctionRule):
    """Return new bounds for cos."""
    func_type = UnaryFunctionType.Cos


class TanRule(_UnboundedFunctionRule):
    """Return new bounds for tan."""
    func_type = UnaryFunctionType.Tan


class AsinRule(_UnboundedFunctionRule):
    """Return new bounds for asin."""
    func_type = UnaryFunctionType.Asin


class AcosRule(_UnboundedFunctionRule):
    """Return new bounds for acos."""
    func_type = UnaryFunctionType.Acos


class AtanRule(_UnboundedFunctionRule):
    """Return new bounds for atan."""
    func_type = UnaryFunctionType.Atan


_func_name_to_rule_map = dict()
_func_name_to_rule_map['abs'] = AbsRule()
_func_name_to_rule_map['sqrt'] = SqrtRule()
_func_name_to_rule_map['exp'] = ExpRule()
_func_name_to_rule_map['log'] = LogRule()
_func_name_to_rule_map['sin'] = SinRule()
_func_name_to_rule_map['cos'] = CosRule()
_func_name_to_rule_map['tan'] = TanRule()
_func_name_to_rule_map['asin'] = AsinRule()
_func_name_to_rule_map['acos'] = AcosRule()
_func_name_to_rule_map['atan'] = AtanRule()


class UnaryFunctionRule(Rule):
    """Bound tightening rule for unary functions."""
    def apply(self, expr, bounds):
        assert len(expr.args) == 1
        func_name = expr.getname()
        if func_name not in _func_name_to_rule_map:
            raise ValueError('Unknown function type {}'.format(func_name))
        return _func_name_to_rule_map[func_name].apply(expr, bounds)



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from tests.strategies import models
from hypothesis import given, reproduce_failure
import pyomo.environ as aml
from suspect.pyomo.expressions import (
    nonpyomo_leaf_types,
    NumericConstant,
)
import numpy as np
from suspect.dag import ProblemDag
from suspect.pyomo.convert import ComponentFactory, dag_from_pyomo_model
from suspect.pyomo.util import model_variables, model_constraints, model_objectives
from suspect.math.arbitrary_precision import inf


def expression_equal(expected, actual):
    if isinstance(actual, NumericConstant):
        assert isinstance(expected, float)
        return np.isclose(expected, actual.value)
    if type(expected).__name__ != type(actual).__name__:
        return False
    if type(expected) in nonpyomo_leaf_types:
        return expected == actual
    if expected.is_variable_type():
        return actual.is_variable_type()
    if len(expected.args) != len(actual.args):
        return False
    for expected_arg, actual_arg in zip(expected.args, actual.args):
        if not expression_equal(expected_arg, actual_arg):
            return False
    return True


class TestConvertVariable(object):
    def test_continuous_variables(self):
        m = aml.ConcreteModel()
        # 10 continuous variables in [-inf, inf]
        m.x = aml.Var(range(10))

        dag = ProblemDag()
        factory = ComponentFactory(dag)
        count = 0
        for omo_var in model_variables(m):
            new_var = factory.variable(omo_var)
            assert new_var.name.startswith('x')
            assert new_var.name == omo_var.name
            assert new_var.lb is None
            assert new_var.ub is None
            assert not new_var.is_integer()
            assert not new_var.is_binary()
            count += 1
        assert count == 10

    def test_integer_variables(self):
        m = aml.ConcreteModel()
        # 5 integer variables in [-10, 5]
        m.y = aml.Var(range(5), bounds=(-10, 5), domain=aml.Integers)

        dag = ProblemDag()
        factory = ComponentFactory(dag)
        count = 0
        for omo_var in model_variables(m):
            new_var = factory.variable(omo_var)
            assert new_var.name.startswith('y')
            assert new_var.lb == -10
            assert new_var.ub == 5
            assert new_var.is_integer()
            count += 1
        assert count == 5

    def test_binary_variables(self):
        m = aml.ConcreteModel()
        # 10 binary variables
        m.b = aml.Var(range(10), domain=aml.Binary)

        dag = ProblemDag()
        factory = ComponentFactory(dag)
        count = 0
        for omo_var in model_variables(m):
            new_var = factory.variable(omo_var)
            assert new_var.name.startswith('b')
            assert new_var.lb == 0
            assert new_var.ub == 1
            assert new_var.is_binary()
            count += 1
        assert count == 10


@given(models(max_constraints=1))
def test_conversion(model):
    dag = dag_from_pyomo_model(model)

    for obj in dag.objectives.values():
        expected_obj = getattr(model, obj.name)
        expected_expr = expected_obj.expr
        assert expression_equal(expected_expr, obj.body)
        print('Done')

    for cons in dag.constraints.values():
        expected_cons = getattr(model, cons.name)
        expected_expr = expected_cons.body
        assert expression_equal(expected_expr, cons.body)
